/* General Purpose */

.clickable:hover {
    cursor: pointer;
}

.color-square { width: 17.4px; height: 17.4px; }

/* Main */

body {
    overscroll-behavior: none;
}

#controls {
    position: fixed;
    top: 56px;
    left: 0;
    box-sizing: border-box;
    width: 260px;
    height: calc(100% - 56px);
    padding-top: 8px;
    padding-left: 8px;
    overflow-y: auto;
    overflow-x: hidden;
}

#zoomControl { width: 100%; margin-bottom: 10px; }
#zoomControl > div { width: 100%; text-align: center; margin-bottom: 4px; }

.control-panel-table { font-size: 0.8em; }
.control-panel-table th { font-size: 0.9em; }

div.bootstrap-select.bs-width-correction > div.dropdown-menu { width: 100%; }
div.bootstrap-select > div.dropdown-menu ul.dropdown-menu { font-size: 0.875rem !important; }
div.bootstrap-select > div.dropdown-menu a.dropdown-item { padding: 4px 8px; }

#assignmentPanelArea { display: none; }
.assignment-panel {
    box-sizing: border-box;
    width: 100%;
    border: 1px solid #ccc;
    padding: 8px;
}
.assignment-panel > h6 { font-size: 0.9em; }
.assignment-panel .btn-group button { font-size: 0.7em; }
p.assignment-panel-bottom-label { font-size: 0.6em; margin-top: 0; font-style: italic; text-align: right; padding-right: 2px; }

#graphs {
    box-sizing: border-box;
    /*padding-left: 270px;*/
    /*height: 100%;*/
    margin-bottom: 100px;



    position: fixed;
    top: 56px;
    left: 270px;
    width: calc(100% - 270px);
    height: calc(100% - 56px);
    padding: 8px 4px 60px 8px;
    overflow-y: auto;
    overflow-x: hidden;
}

#metadata {
    width: 100%;
    margin-bottom: 14px;
}

/* Matching dygraphs headers for now */
#metadata > h2 {
    font-size: 20px;
    text-align: center;
    font-weight: bold;
}

.graph_wrapper {
    width: 100%;
    height: 120px;
}

.graph_wrapper > table {
    width: 100%;
    height: 100%;
    table-layout: fixed;
}

.graph_wrapper > table > tbody > tr > td.graph_title, .graph_wrapper > table > tbody > tr > td.legend {
    width: 120px;
    padding-left: 4px;
}

.graph_wrapper > table > tbody > tr > td.graph_title {
    vertical-align: bottom;
    overflow: hidden;
    height: 60px;
    padding-bottom: 4px;
    font-weight: bold;
}

.graph_wrapper > table > tbody > tr > td.graph_title > .graph_subtitle {
    font-weight: normal;
    font-size: .8em;
    margin-top: 8px;
    word-wrap: break-word;
}

.graph_wrapper > table > tbody > tr > td.graph_title > .webix_icon {
    color: rgb(71, 84, 102);
    visibility: hidden;
    cursor: pointer;
    margin-left: 8px;
}

.graph_wrapper > table > tbody > tr:hover > td.graph_title > .webix_icon {
    visibility: visible !important;
}

.graph_wrapper > table > tbody > tr > td.legend {
    vertical-align: top;
    /*height: 120px;*/
    padding-top: 4px;;
}

.graph_wrapper > table > tbody > tr > td.legend > div {
    position: relative;
    display: inline-block;
    background-color: white;
    filter: drop-shadow(-2px 2px 2px #ccc);
    border: 1px solid #aaa;
    padding: 2px 4px;
    word-wrap: break-word;
    z-index: 99999;
}

.graph_wrapper > table > tbody > tr > td.legend > div > span {
    white-space: nowrap;
}

.graph {
    width: 100%;
    /*height: 120px;*/
    height: 100%;
}

/* Realtime */

body.realtime > #controls > #projectAndFileSelection, body.realtime > #controls > #patternDetectionPanel {
    display: none !important;
}
body.realtime #realtimeTitle {
    /*display: initial !important;*/
    display: inline !important;
}

#control_footer {
    width: 100%;
    clear: both;
    text-align: center;
    margin-top: 30px;
}

#control_footer > p#enterrealtime, #control_footer > p#exitrealtime {
    cursor: pointer;
    color: #888;
    font-size: 80%;
}

body:not(.realtime) > #controls > #control_footer > p#enterrealtime {
    display: block;
}

body:not(.realtime) > #controls > #control_footer > p#exitrealtime {
    display: none !important;
}

body.realtime > #controls > #control_footer > p#enterrealtime {
    display: none !important;
}

body.realtime > #controls > #control_footer > p#exitrealtime {
    display: block;
}

/* Annotations */

#allAnnotationsListModal tr:hover, #annotationsListModal tr:hover {
    cursor: pointer;
}

#annotationStartDate, #annotationStartTime, #annotationEndDate, #annotationEndTime {
    width: 85%;
    margin-bottom: 3px;
}

.webix_message_area {
    display: none !important;
}

#interactionsModal table {
    width: 100%;
}

#interactionsModal table > tbody > tr > td:first-child {
    padding-right: 24px;
}